import http.client
import sre_compile
import time
import json
import http.client
from typing import *

import yaml
try:
    from config import NetworkConfig
    TIMEOUT = NetworkConfig.TIMEOUT
except ImportError:
    TIMEOUT = 15

from .gtlog import GTLog
from .expression import ExpressionParser
from .thirdparty_packages import requests


class PocArgs(object):
    """
    poc 规则单个 uri 参数
    """
    def __init__(self, jdata: Dict) -> None:
        """
        构造
        """
        self.uri = jdata.get('uri', '')
        self.method = jdata.get('method', 'GET')
        self.headers = jdata.get('headers', {})
        self.body: str = jdata.get('body', '')
        self.allow_redirects: bool = jdata.get("allow_redirects", True)
        self.expression: str = str(jdata.get('expression', ''))
        self.timeout = jdata.get('timeout', TIMEOUT)

    def __repr__(self):
        return str(self.__dict__)


class PocVars(object):
    """
    poc 规则运行时，对指定网页运行产生的结果
    """
    def __init__(self, poc_rule: 'PocRule') -> None:
        """
        构造
        """

        # 规则、返回包
        self.poc_rule = poc_rule

        # 记录运行时信息
        self.start_time: List[float] = []  # 每个rule运行时的开始时间
        self.use_time: List[float] = []  # 每个rule运行的时间
        self.request = []
        self.response = []

    def replace_vars(self, args, parser: ExpressionParser):
        if isinstance(args, str):
            start = args.find("{{")
            end = args.find("}}")
        elif isinstance(args, bytes):
            start = args.find(b"{{")
            end = args.find(b"}}")
        else:
            return args
        if end > start > -1:
            arg_name = args[start + 2: end]  # 提取表达式
            flag = "{{" + arg_name + "}}"  # 要替换的标志位
            var = parser.parse(arg_name)  # 解析表达式
            if isinstance(var, bytes):
                args = args.encode().replace(flag.encode(), var)
            else:
                args = args.replace(flag, str(var))
            return self.replace_vars(args, parser)
        return args

    def network_start(self, poc_args: PocArgs, parser: ExpressionParser) -> None:
        """
        网络操作开始, 记录开始时间，使用variables, group替换poc_args中的{{}}变量
        """
        poc_args.uri = self.replace_vars(poc_args.uri, parser)
        poc_args.body = self.replace_vars(poc_args.body, parser)
        for k in poc_args.headers:
            poc_args.headers[k] = self.replace_vars(poc_args.headers[k], parser)
        self.start_time.append(time.time())

    def network_done(self, resp) -> None:
        """
        网络操作完成，记录请求响应信息
        """
        self.request.append(resp.request)
        self.response.append(resp)

        self.use_time.append(time.time() - self.start_time[-1])


class PocRule(object):  # pylint: disable=too-many-instance-attributes
    """
    从路径初始化poc 规则
    """
    def __init__(self, relativepath: str, debug=False, proxies=None) -> None:
        """
        构造
        """
        self.debug = debug
        path = relativepath.replace(".", "/")
        path = path if path.endswith(".yml") else path + ".yml"
        with open(path, encoding="utf-8") as f:
            jdata = yaml.safe_load(f)

        if debug:
            self.log("Plugin load success:")
            self.log(jdata)

        # 信息参数
        self.product = jdata.get("product", '')
        self.name = jdata.get('name', '')
        self.holetype = jdata.get('holetype', 'other')
        self.level = jdata.get('level', None)

        self.desc = jdata.get('desc', '')
        self.author = jdata.get('author', '')
        self.cve = jdata.get('cve', None)
        self.cnvd = jdata.get('cnvd', None)

        # 规则参数
        self.poc_args = [PocArgs(poc_args) for poc_args in jdata.get('rules', [])]

        # 可选参数
        self.control = jdata.get('control', 'SEQ')  # 多个rule之间逻辑 SEQ/AND/OR
        self.http_version = jdata.get('http_version', "1.1")  # HTTP版本
        http.client.HTTPConnection._http_vsn_str = f"HTTP/{self.http_version}"
        self.set = jdata.get('set', {})  # 变量预定义
        self.ret = jdata.get('ret', f'检测到 {self.name}, {self.desc}')  # 命中后返回内容

        # 表达式解析器
        self._parser = None

        # 记录poc运行时的产生的结果
        self.poc_vars = PocVars(self)
        self.proxies = proxies if proxies else {}
        self.session = requests.Session()

        # 初始化dnslog
        self.gtlog = GTLog()

    def __repr__(self) -> str:
        return str(self.__dict__)

    @property
    def parser(self):
        if not self._parser:
            self._parser = ExpressionParser(proxies=self.proxies)
        return self._parser

    def init_set(self):
        # 初始化poc中的set变量
        for k, expression in self.set.items():
            value = self._parser.parse(expression)
            self.parser.add_variable(**{k: value})

    def execute(self, url):
        result = {"status": False}
        # 初始化运行过程中变量
        self.parser.add_variable(url=url)
        self.init_set()

        if self.debug and self.proxies:
            self.log(f"正在使用代理: {self.proxies}")

        self.poc_vars = PocVars(self)
        expression_results = []
        for i, poc_args in enumerate(self.poc_args):
            if poc_args.uri:
                self.poc_vars.network_start(poc_args, self.parser)
                resp = self.session.request(
                    method=poc_args.method,
                    url=f'{url}{poc_args.uri}',
                    headers=poc_args.headers,
                    data=poc_args.body,
                    allow_redirects=poc_args.allow_redirects,
                    timeout=poc_args.timeout,
                    proxies=self.proxies
                )
                self.poc_vars.network_done(resp)
                self.parser.response = resp
                if self.debug:
                    self.log(f"第{i + 1}次请求:")
                    self.log(requests.dump_all(resp))
            exp_result = self.__parse_expression(poc_args.expression)

            if self.debug:
                self.log(f"函数解析结果: {self.parser.last_ret}")
                self.log(f"表达式解析结果: {exp_result}")
                self.log(f"当前可用变量：{self.parser.variables}")

            expression_results.append(exp_result)
            if self.control == "OR" and exp_result is True:
                # OR规则只需要有一个为真即跳出检测
                break
            if self.control != "OR" and exp_result is False:
                # AND, SEQ规则只需要有一个为假即跳出检测
                break

        # 判断是否命中
        if self.control == "OR":
            result["status"] = any(expression_results)
        else:
            result["status"] = all(expression_results)

        if result["status"]:
            result["ret"] = self.poc_vars.replace_vars(self.ret, self.parser)
        if self.debug:
            self.log("*" * 10)
            self.log(f"检测结果：{result}")
        return result

    def __parse_expression(self, expression: str) -> bool:
        # 解析rule表达式，并记录正则匹配结果作为新的变量
        parse_ret = self.parser.parse(expression)
        for ret in self.parser.last_ret:
            if isinstance(ret, type(sre_compile.compile('', 0).match(''))):
                self.parser.add_variable(**ret.groupdict())  # 按分组匹配结果变量存储 e.g. (?P<username>\w+) username=
        return bool(parse_ret)

    def log(self, *args):
        for arg in args:
            if isinstance(arg, dict):
                print(json.dumps(arg, indent=2))
            else:
                print(arg)


if __name__ == "__main__":
    rule = PocRule("plugins.poc_yml.rails_cve_2018_3760_rce", debug=True)
    print(rule.execute("http://10.91.4.101:3000/"))
    print(rule.parser.variables)
    rule.log()
